/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.apphelpers;

import pl.chyla.wyjec.apphelpers.FixedSizeLRUCache.CacheEntry;

import junit.framework.TestCase;


/**
 * Tests for {@link FixedSizeLRUCache} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestFixedSizeLRUCache extends TestCase {

    /**
     * Empty default constructor
     */
    public TestFixedSizeLRUCache() { }

    /**
     * Checks whether given cache entry represents specified key-value pair.
     *
     * @param entry cache entry to check
     * @param expKey expected key
     * @param expValue expected value
     */
    private void checkEntry(CacheEntry entry, Object expKey, Object expValue) {
        assertEquals(entry.getKey(), expKey);
        assertEquals(entry.getValue(), expValue);
    }

    /**
     * Tests {@link FixedSizeLRUCache#LRUCache(int)} constructor.
     */
    public void testConstructor() {
        new FixedSizeLRUCache(1);
        new FixedSizeLRUCache(1000);
        try {
            new FixedSizeLRUCache(0);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            new FixedSizeLRUCache(-1);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

    /**
     * Tests if {@link FixedSizeLRUCache} behaves correctly when trying to exceed its size.
     */
    public void testSizeLimit() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(1);
        cache.add("k-1", "v-1");
        try {
            cache.add("k-2", "v-2");
            fail();
        } catch (FixedSizeLRUCache.SizeExceededException e) {
            // OK
        }
        checkEntry(cache.removeLeastRecentlyUsed(), "k-1", "v-1");
        cache.add("k-2", "v-2");
        cache = new FixedSizeLRUCache(3);
        cache.add("k-1", "v-1");
        cache.add("k-2", "v-2");
        cache.add("k-3", "v-3");
        try {
            cache.add("k-4", "v-4");
            fail();
        } catch (FixedSizeLRUCache.SizeExceededException e) {
            // OK
        }
        checkEntry(cache.removeLeastRecentlyUsed(), "k-1", "v-1");
        checkEntry(cache.removeLeastRecentlyUsed(), "k-2", "v-2");
        checkEntry(cache.removeLeastRecentlyUsed(), "k-3", "v-3");
        cache.add("k-1", "v-1");
        cache.add("k-2", "v-2");
        cache.add("k-3", "v-3");
        try {
            cache.add("k-4", "v-4");
            fail();
        } catch (FixedSizeLRUCache.SizeExceededException e) {
            // OK
        }
    }

    /**
     * Tests if {@link FixedSizeLRUCache} behaves correctly when trying to exceed its size
     * (using {@link StaticHashObject} instances as keys).
     */
    public void testSizeLimitStaticHash() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(1);
        cache.add(new StaticHashObject("k-1"), "v-1");
        try {
            cache.add(new StaticHashObject("k-2"), "v-2");
            fail();
        } catch (FixedSizeLRUCache.SizeExceededException e) {
            // OK
        }
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-1"), "v-1");
        cache.add(new StaticHashObject("k-2"), "v-2");
        cache = new FixedSizeLRUCache(3);
        cache.add(new StaticHashObject("k-1"), "v-1");
        cache.add(new StaticHashObject("k-2"), "v-2");
        cache.add(new StaticHashObject("k-3"), "v-3");
        try {
            cache.add(new StaticHashObject("k-4"), "v-4");
            fail();
        } catch (FixedSizeLRUCache.SizeExceededException e) {
            // OK
        }
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-1"), "v-1");
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-2"), "v-2");
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-3"), "v-3");
        cache.add(new StaticHashObject("k-1"), "v-1");
        cache.add(new StaticHashObject("k-2"), "v-2");
        cache.add(new StaticHashObject("k-3"), "v-3");
        try {
            cache.add(new StaticHashObject("k-4"), "v-4");
            fail();
        } catch (FixedSizeLRUCache.SizeExceededException e) {
            // OK
        }
    }

    /**
     * Tests {@link FixedSizeLRUCache#add(Object, Object)} and
     * {@link FixedSizeLRUCache#get(Object)} methods.
     */
    public void testAddGet() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(5);
        assertEquals(cache.get("k-1"), null);
        cache.add("k-1", "v-1");
        assertEquals(cache.get("k-1"), "v-1");
        assertEquals(cache.get("k-2"), null);
        assertEquals(cache.get("k-3"), null);
        cache.add("k-2", "v-2");
        assertEquals(cache.get("k-1"), "v-1");
        assertEquals(cache.get("k-2"), "v-2");
        assertEquals(cache.get("k-3"), null);
        cache.add("k-3", "v-3");
        assertEquals(cache.get("k-1"), "v-1");
        assertEquals(cache.get("k-2"), "v-2");
        assertEquals(cache.get("k-3"), "v-3");
    }

    /**
     * Tests {@link FixedSizeLRUCache#add(Object, Object)} and
     * {@link FixedSizeLRUCache#get(Object)} methods (using {@link StaticHashObject} instances
     * as keys).
     */
    public void testAddGetStaticHash() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(5);
        assertEquals(cache.get(new StaticHashObject("k-1")), null);
        cache.add(new StaticHashObject("k-1"), "v-1");
        assertEquals(cache.get(new StaticHashObject("k-1")), "v-1");
        assertEquals(cache.get(new StaticHashObject("k-2")), null);
        assertEquals(cache.get(new StaticHashObject("k-3")), null);
        cache.add(new StaticHashObject("k-2"), "v-2");
        assertEquals(cache.get(new StaticHashObject("k-1")), "v-1");
        assertEquals(cache.get(new StaticHashObject("k-2")), "v-2");
        assertEquals(cache.get(new StaticHashObject("k-3")), null);
        cache.add(new StaticHashObject("k-3"), "v-3");
        assertEquals(cache.get(new StaticHashObject("k-1")), "v-1");
        assertEquals(cache.get(new StaticHashObject("k-2")), "v-2");
        assertEquals(cache.get(new StaticHashObject("k-3")), "v-3");
    }

    /**
     * Tests {@link FixedSizeLRUCache#add(Object, Object)} method by trying to add the same key
     * twice.
     */
    public void testDoubleAdd() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(5);
        cache.add("k-1", "v-1");
        cache.add("k-2", "v-2");
        cache.add("k-3", "v-3");
        try {
            cache.add("k-2", "v-2");
            fail();
        } catch (FixedSizeLRUCache.DuplicateKeyException e) {
            // OK
        }
        cache.get("k-1");
        cache.get("k-3");
        checkEntry(cache.removeLeastRecentlyUsed(), "k-2", "v-2");
        cache.add("k-2", "v-2");
    }

    /**
     * Tests {@link FixedSizeLRUCache#add(Object, Object)} method by trying to add the same key
     * twice (using {@link StaticHashObject} instances as keys).
     */
    public void testDoubleAddStatichash() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(5);
        cache.add(new StaticHashObject("k-1"), "v-1");
        cache.add(new StaticHashObject("k-2"), "v-2");
        cache.add(new StaticHashObject("k-3"), "v-3");
        try {
            cache.add(new StaticHashObject("k-2"), "v-2");
            fail();
        } catch (FixedSizeLRUCache.DuplicateKeyException e) {
            // OK
        }
        cache.get(new StaticHashObject("k-1"));
        cache.get(new StaticHashObject("k-3"));
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-2"), "v-2");
        cache.add(new StaticHashObject("k-2"), "v-2");
    }

    /**
     * Tests {@link FixedSizeLRUCache#hasFreeSpace()} method.
     */
    public void testHasFreeSpace() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(1);
        assertEquals(cache.hasFreeSpace(), true);
        cache.add("k-1", "v-1");
        assertEquals(cache.hasFreeSpace(), false);
        checkEntry(cache.removeLeastRecentlyUsed(), "k-1", "v-1");
        assertEquals(cache.hasFreeSpace(), true);
        cache = new FixedSizeLRUCache(3);
        assertEquals(cache.hasFreeSpace(), true);
        cache.add("k-1", "v-1");
        cache.add("k-2", "v-2");
        assertEquals(cache.hasFreeSpace(), true);
        cache.add("k-3", "v-3");
        assertEquals(cache.hasFreeSpace(), false);
        checkEntry(cache.removeLeastRecentlyUsed(), "k-1", "v-1");
        assertEquals(cache.hasFreeSpace(), true);
        cache.add("k-4", "v-4");
        assertEquals(cache.hasFreeSpace(), false);
    }

    /**
     * Tests {@link FixedSizeLRUCache#hasFreeSpace()} method (using {@link StaticHashObject}
     * instances as keys).
     */
    public void testHasFreeSpaceStaticHash() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(1);
        assertEquals(cache.hasFreeSpace(), true);
        cache.add(new StaticHashObject("k-1"), "v-1");
        assertEquals(cache.hasFreeSpace(), false);
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-1"), "v-1");
        assertEquals(cache.hasFreeSpace(), true);
        cache = new FixedSizeLRUCache(3);
        assertEquals(cache.hasFreeSpace(), true);
        cache.add(new StaticHashObject("k-1"), "v-1");
        cache.add(new StaticHashObject("k-2"), "v-2");
        assertEquals(cache.hasFreeSpace(), true);
        cache.add(new StaticHashObject("k-3"), "v-3");
        assertEquals(cache.hasFreeSpace(), false);
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-1"), "v-1");
        assertEquals(cache.hasFreeSpace(), true);
        cache.add(new StaticHashObject("k-4"), "v-4");
        assertEquals(cache.hasFreeSpace(), false);
    }

    /**
     * Tests {@link FixedSizeLRUCache#removeLeastRecentlyUsed()} method.
     */
    public void testRemoveLeastRecentlyUsed() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(1);
        cache.add("k-1", "v-1");
        checkEntry(cache.removeLeastRecentlyUsed(), "k-1", "v-1");
        assertEquals(cache.get("k-1"), null);
        cache = new FixedSizeLRUCache(3);
        cache.add("k-1", "v-1");
        cache.add("k-2", "v-2");
        cache.add("k-3", "v-3");
        checkEntry(cache.removeLeastRecentlyUsed(), "k-1", "v-1");
        assertEquals(cache.get("k-1"), null);
        checkEntry(cache.removeLeastRecentlyUsed(), "k-2", "v-2");
        assertEquals(cache.get("k-2"), null);
        cache.add("k-4", "v-4");
        cache.add("k-5", "v-5");
        cache.get("k-3");
        checkEntry(cache.removeLeastRecentlyUsed(), "k-4", "v-4");
        assertEquals(cache.get("k-4"), null);
        checkEntry(cache.removeLeastRecentlyUsed(), "k-5", "v-5");
        assertEquals(cache.get("k-5"), null);
        checkEntry(cache.removeLeastRecentlyUsed(), "k-3", "v-3");
        assertEquals(cache.get("k-3"), null);
    }

    /**
     * Tests {@link FixedSizeLRUCache#removeLeastRecentlyUsed()} method (using
     * {@link StaticHashObject} instances as keys).
     */
    public void testRemoveLeastRecentlyUsedStaticHash() {
        FixedSizeLRUCache cache = new FixedSizeLRUCache(1);
        cache.add(new StaticHashObject("k-1"), "v-1");
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-1"), "v-1");
        assertEquals(cache.get(new StaticHashObject("k-1")), null);
        cache = new FixedSizeLRUCache(3);
        cache.add(new StaticHashObject("k-1"), "v-1");
        cache.add(new StaticHashObject("k-2"), "v-2");
        cache.add(new StaticHashObject("k-3"), "v-3");
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-1"), "v-1");
        assertEquals(cache.get(new StaticHashObject("k-1")), null);
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-2"), "v-2");
        assertEquals(cache.get(new StaticHashObject("k-2")), null);
        cache.add(new StaticHashObject("k-4"), "v-4");
        cache.add(new StaticHashObject("k-5"), "v-5");
        cache.get(new StaticHashObject("k-3"));
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-4"), "v-4");
        assertEquals(cache.get(new StaticHashObject("k-4")), null);
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-5"), "v-5");
        assertEquals(cache.get(new StaticHashObject("k-5")), null);
        checkEntry(cache.removeLeastRecentlyUsed(), new StaticHashObject("k-3"), "v-3");
        assertEquals(cache.get(new StaticHashObject("k-3")), null);
    }


    /**
     * Simple named object with <code>HashCode</code> implementation that returns a constant
     * value.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private static class StaticHashObject {

        /**
         * This objects's name.
         */
        private final String name;

        /**
         * Constructs <code>StaticHashObject</code> instance with given name.
         *
         * @param name object's name
         */
        StaticHashObject(String name) {
            this.name = name;
        }

        /**
         * Checks if this object is equal to another one.
         *
         * @return
         *     <code>true</code> if <code>other</code> is <code>StaticHashObject</code> and both
         *     instances have equal names, <code>false</code> otherwise
         */
        public boolean equals(Object other) {
            if (!(other instanceof StaticHashObject)) {
                return false;
            }
            return name.equals(((StaticHashObject) other).name);
        }

        /**
         * Returns a hash code value for the object, using always the same value.
         *
         * @return constant value
         */
        public int hashCode() {
            return 1;
        }

    }

}
