/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.apphelpers;

import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.parser.ExpressionTree;
import pl.chyla.wyjec.parser.InvalidInputException;
import pl.chyla.wyjec.parser.Parser;
import pl.chyla.wyjec.parser.TestAtomNode;
import pl.chyla.wyjec.stdconsts.StandardConstantsParser;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.stdfuncs.integerfuncs.IntegerFunctions;
import pl.chyla.wyjec.utils.RawString;


/**
 * Tests for {@link SimpleParser} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestSimpleParser extends TestCase {

    /**
     * Empty default constructor
     */
    public TestSimpleParser() { }

    /**
     * Compares results of parsing given expression by two specified parsers. Reports assertion
     * exception in case the results are different. The expression may be valid (in that case
     * both parsers should return equivalent expression trees) or it may contain errors (both
     * parsers should report an error).
     *
     * @param expression expression to use for comparing parser
     * @param parser raw parser to use
     * @param simpleParser simple parser to compare with the raw one
     */
    private void checkParsersCompatibility(
        String expression, Parser parser, SimpleParser simpleParser) {
        ExpressionTree aExprTree;
        ExpressionTree bExprTree;
        try {
            aExprTree = parser.parse(new RawString(expression));
        } catch (InvalidInputException e) {
            aExprTree = null;
        }
        try {
            bExprTree = simpleParser.parse(expression);
        } catch (InvalidInputException e1) {
            bExprTree = null;
        }
        if (aExprTree != null && bExprTree != null) {
            TestAtomNode.checkAtomNodeEquivalence(
                aExprTree.getTopLevelAtom(), bExprTree.getTopLevelAtom());
        } else {
            assertTrue(aExprTree == null && bExprTree == null);
        }
    }

    /**
     * Tests {@link SimpleParser#parse(String)} method by comparing its results with results
     * of {@link Parser#parse(ReadableRawString)}.
     */
    public void testParse() {
        String[] expressions = {
            "", " ", "Test: {5}", "Test: {+|{5}|{10}}", "Test: {upper|abc}!", "Test: {}!", "}",
            "{", "{+", "{+|", "{+|a", "{+|a}", "{hmm|x}"};
        Config config = Config.getDefaultConfig();
        Function[] funcs = StandardFunctions.getFunctions();
        IConstantsParser constsParser = StandardConstantsParser.getInstance();
        Parser parser = new Parser(config, funcs, constsParser);
        SimpleParser simpleParser = new SimpleParser(config, funcs, constsParser);
        for (int i = 0; i < expressions.length; i++) {
            checkParsersCompatibility(expressions[i], parser, simpleParser);
        }
        simpleParser = new SimpleParser();
        for (int i = 0; i < expressions.length; i++) {
            checkParsersCompatibility(expressions[i], parser, simpleParser);
        }
        config = Config.getZeroMemLimitsConfig();
        funcs = IntegerFunctions.getFunctions();
        parser = new Parser(config, funcs, constsParser);
        simpleParser = new SimpleParser(config, funcs, constsParser);
        for (int i = 0; i < expressions.length; i++) {
            checkParsersCompatibility(expressions[i], parser, simpleParser);
        }
    }

}
