/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.core;

import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.utils.RawString;


/**
 * Base class for classes testing Wyjec functions. Contains helper methods for creating
 * {@link Value} instances and method for evaluating specified function with given parameters.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class BaseTestFunction extends TestCase {

    /**
     * Default soft limit for a number of actual parameters, for use as a "resonably big" number
     * of parameters in tests.
     */
    protected static final int DEFAULT_PARAMS_SOFT_LIMIT =
        Config.getDefaultConfig().getFuncActualParamsSoftLimit();

    /**
     * Length of storage in {@link RawString} instance used for storing results of evaluation
     * of string functions.
     */
    private int resultStringLength;

    /**
     * Checks whether given sequence of types of actual parameters matches declaration
     * consisting of given sequence of types of required parameters and type of optional
     * parameters.
     *
     * @param reqParamsTypes array of types of declared required parameters
     * @param optParamsType
     *     declared type of optional parameters or {@link Type#NONE} if a declaration doesn't
     *     allow optional parameters
     * @param actParamsTypes types of actual parameters to match to declaration
     * @return
     *     <code>true</code> if given actual parameters types match specified declaration,
     *     <code>false</code> otherwise
     */
    public static boolean checkParametersTypesMatch(
        Type[] reqParamsTypes, Type optParamsType, Type[] actParamsTypes) {
        boolean isMatch = true;
        isMatch &= actParamsTypes.length >= reqParamsTypes.length;
        for (int i = 0; i < reqParamsTypes.length && isMatch; i++) {
            isMatch &= actParamsTypes[i] == reqParamsTypes[i];
        }
        if (optParamsType != Type.NONE) {
            for (int i = reqParamsTypes.length; i < actParamsTypes.length && isMatch; i++) {
                isMatch &= actParamsTypes[i] == optParamsType;
            }
        } else {
            isMatch &= actParamsTypes.length == reqParamsTypes.length;
        }
        return isMatch;
    }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        resultStringLength = Config.getDefaultConfig().getStringLengthLimit();
    }

    /**
     * Returns function instance used by this test case for evaluation
     *
     * @return function used by this test case
     */
    protected abstract Function getFunction();

    /**
     * Equivalent for {@link TestValue#makeBoolean(boolean)} method, defined to save typing
     * inside functions tests.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    protected final Value makeBooleanValue(boolean value) {
        return TestValue.makeBoolean(value);
    }

    /**
     * Equivalent for {@link TestValue#makeInteger(long)} method, defined to save typing inside
     * functions tests.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    protected final Value makeIntegerValue(long value) {
        return TestValue.makeInteger(value);
    }

    /**
     * Equivalent for {@link TestValue#makeString(String)} method, defined to save typing inside
     * functions tests.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    protected final Value makeStringValue(String value) {
        return TestValue.makeString(value);
    }

    /**
     * Equivalent for {@link TestValue#makeStringVariants(String)} method, defined to save
     * typing inside functions tests.
     *
     * @param value value to build {@link Value} instances for
     * @return a list of generated instances
     */
    protected final Value[] makeStringValueVariants(String value) {
        return TestValue.makeStringVariants(value);
    }

    /**
     * Equivalent for {@link TestValue#makeFloat(double)} method, defined to save typing inside
     * functions tests.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    protected final Value makeFloatValue(double value) {
        return TestValue.makeFloat(value);
    }

    /**
     * Equivalent for {@link TestValue#makeDate(long)} method, defined to save typing inside
     * functions tests.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    protected final Value makeDateValue(long value) {
        return TestValue.makeDate(value);
    }

    /**
     * Equivalent for {@link TestValue#makeDate(int, int, int, int, int, int)} method, defined
     * to save typing inside functions tests.
     *
     * @param year year number
     * @param month 1-based month of the year
     * @param day 1-based day of the month
     * @param hour 0-based hour of the day
     * @param minute 0-based minute of the hour
     * @param second 0-based second of the minute
     * @return newly created {@link Value} instance
     */
    protected final Value makeDateValue(
        int year, int month, int day, int hour, int minute, int second) {
        return TestValue.makeDate(year, month, day, hour, minute, second);
    }

    /**
     * Tests whether every call of {@link Function#getSignature()} evaluates the same signature
     * instance.
     */
    public void testSameSignature() {
        Function func = getFunction();
        assertSame(func.getSignature(), func.getSignature());
    }

    /**
     * Tests whether every call of {@link Function#getName()} evaluates equal name.
     */
    public void testEqualName() {
        Function func = getFunction();
        assertEquals(func.getName(), func.getName());
    }

    /**
     * Sets length of string used for storing results of string functions.
     *
     * @param length new length of string used for string functions results
     */
    protected final void setResultStringLength(int length) {
        resultStringLength = length;
    }

    /**
     * Returns length of string used for storing results of string functions.
     *
     * @return length of string used for string functions results
     */
    protected final int getResultStringLength() {
        return resultStringLength;
    }

    /**
     * Evaluates function used by this test case with given parameters and returns {@link Value}
     * instance containing the result.
     * @param params function's actual parameters
     *
     * @return result of evaluation
     */
    protected final Value rawEvaluateFunction(Value[] params) {
        Function func = getFunction();
        checkParametersAgainstSignature(params, func.getSignature());
        Value result = new Value(resultStringLength);
        func.evaluate(new EvaluationContext(result, params));
        return result;
    }

    /**
     * Checks if given array of parameters matches given signature, throwing assertion exception
     * otherwise.
     *
     * @param params parameters to checks against the signature
     * @param signature signature to use
     */
    private void checkParametersAgainstSignature(Value[] params, FunctionSignature signature) {
        Type[] paramsTypes = new Type[params.length];
        for (int i = 0; i < params.length; i++) {
            paramsTypes[i] = params[i].getType();
        }
        assertTrue(checkParametersTypesMatch(
            signature.getRequiredParametersTypes(), signature.getOptionalParametersType(),
            paramsTypes));
    }


    /**
     * Simple implementation of evaluation context, providing function parameters directly from
     * array of precalculated {@link Value} instances.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private class EvaluationContext implements IEvaluationContext {

        /**
         * {@link Value} instance for storing the result of evaluation
         */
        private final Value result;

        /**
         * Values of function parameters
         */
        private final Value[] params;

        /**
         * Constructs <code>EvaluationContext</code> instance using given parameters and result
         * object.
         *
         * @param result object for storing evaluation result
         * @param params parameters for function evaluation
         */
        EvaluationContext(Value result, Value[] params) {
            this.result = result;
            this.params = params;
        }

        /**
         * {@inheritDoc}
         */
        public int getParametersCount() {
            return params.length;
        }

        /**
         * {@inheritDoc}
         */
        public Value getParameter(int paramIndex) {
            return params[paramIndex];
        }

        /**
         * {@inheritDoc}
         */
        public Value getResult() {
            return result;
        }

    }

}
