/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.core;

import junit.framework.TestCase;


/**
 * Tests for {@link Parameter} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestParameter extends TestCase {

    /**
     * Empty default constructor
     */
    public TestParameter() { }

    /**
     * Creates new instance of anonymous subclass if {@link Parameter} that represents a string
     * parameter.
     *
     * @return new object representing string parameter
     */
    private Parameter makeStringParameter() {
        Parameter param = new Parameter() {
            public Type getType() {
                return Type.STRING;
            }
        };
        return param;
    }

    /**
     * Tests {@link Parameter#configure(int)} method.
     */
    public void testConfigure() {
        Parameter param = makeStringParameter();
        param.configure(2);
    }

    /**
     * Tests {@link Parameter#configure(int)} method trying to call it twice.
     */
    public void testConfigureRepeatet() {
        Parameter param = makeStringParameter();
        param.configure(2);
        try {
            param.configure(2);
            fail();
        } catch (IllegalStateException e) {
            // OK
        }
    }

    /**
     * Tests {@link Parameter#configure(int)} method with invalid <code>paramIndex</code>
     * argument.
     */
    public void testConfigureInvalidArg() {
        Parameter param;
        try {
            param = makeStringParameter();
            param.configure(-1);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            param = makeStringParameter();
            param.configure(Integer.MIN_VALUE);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

    /**
     * Tests {@link Parameter#getIndex()} method.
     */
    public void testGetIndex() {
        Parameter param = makeStringParameter();
        param.configure(2);
        assertEquals(param.getIndex(), 2);
        param = makeStringParameter();
        param.configure(Integer.MAX_VALUE);
        assertEquals(param.getIndex(), Integer.MAX_VALUE);
    }

    /**
     * Tests {@link Parameter#getIndex()} method by calling it on unconfigured object.
     */
    public void testGetIndexUnconfigured() {
        Parameter param = makeStringParameter();
        try {
            param.getIndex();
            fail();
        } catch (IllegalStateException e) {
            // OK
        }
    }

    /**
     * Tests {@link Parameter#getType()} method.
     */
    public void testGetType() {
        Parameter param = makeStringParameter();
        assertSame(param.getType(), Type.STRING);
        param.configure(2);
        assertSame(param.getType(), Type.STRING);
    }

}
