/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.core.functions;

import pl.chyla.wyjec.core.BaseTestFunction;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.RawString;


/**
 * Base class for classes testing string functions (functions returning string values).
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class BaseTestStringFunction extends BaseTestFunction {

    /**
     * Instance of string function used for evaluation during testing.
     */
    private final StringFunction function;

    /**
     * Constructs string function test using given function instance.
     *
     * @param function function to use for evaluation during testing
     */
    protected BaseTestStringFunction(StringFunction function) {
        this.function = function;
    }

    /**
     * {@inheritDoc}
     */
    protected final Function getFunction() {
        return function;
    }

    /**
     * Tests whether the function declares the appropriate return type.
     */
    public void testReturnType() {
        assertSame(function.getResultType(), Type.STRING);
    }

    /**
     * Evaluates function associated with the test, using parameters from given array.
     *
     * @param params parameters to pass to the function
     * @return function's result
     */
    protected final String evaluateFunction(Value[] params) {
        return rawEvaluateFunction(params).getString().toString();
    }

    /**
     * Evaluates function associated with the test using given parameters and checks if the
     * result of evaluation is equal to given expected value. If these values are different,
     * the method fails with assertion exception.
     *
     * @param params parameters to pass to the function
     * @param expResult expected result of function evaluation
     */
    protected final void checkFunctionResult(Value[] params, String expResult) {
        assertEquals(evaluateFunction(params), expResult);
    }

    /**
     * Changes the length of result string a few times, evaluating function associated with the
     * test after every change. After every evaluation it checks if its result is equal to given
     * expected value or {@link RawString.RawStringOverflowException} is thrown when the result
     * string is too small. This method is the simplest way to test well-behaving functions,
     * i.e. such that:
     * <ul>
     * <li>store valid evaluation result in result string if it's big enough</li>
     * <li>throw {@link RawString.RawStringOverflowException} otherwise</li>
     * </li>
     * This method doesn't catch exceptions other than
     * {@link RawString.RawStringOverflowException}. Besides, even this exception won't be
     * caught if the function is not supposed to throw it, i.e. when the result string is big
     * enough for the expected result.
     * <p>
     * This method restores original length of the result string before returning.
     *
     * @param params parameters to pass to the function
     * @param expResult expected result of function evaluations
     */
    protected final void checkFunctionResultVarSizes(Value[] params, String expResult) {
        int oldStringLength = getResultStringLength();
        if (expResult.length() > 0) {
            setResultStringLength(0);
            checkFunctionEvaluationOverflow(params);
        }
        if (expResult.length() > 1) {
            setResultStringLength(expResult.length() - 1);
            checkFunctionEvaluationOverflow(params);
        }
        setResultStringLength(expResult.length());
        assertEquals(evaluateFunction(params), expResult);
        setResultStringLength(expResult.length() + 1);
        assertEquals(evaluateFunction(params), expResult);
        setResultStringLength(oldStringLength);
    }

    /**
     * Evaluates function associated with the test using given parameters and checks if
     * {@link RawString.RawStringOverflowException} exception is thrown during evaluation.
     * This method fails if function evaluation succeeds. Exception other than
     * {@link RawString.RawStringOverflowException} will not be catched. 
     *
     * @param params parameters to pass to the function
     */
    protected final void checkFunctionEvaluationOverflow(Value[] params) {
        try {
            evaluateFunction(params);
            fail();
        } catch (RawString.RawStringOverflowException e) {
            // OK
        }
    }

}
