/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.evaluator;

import java.util.Arrays;
import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.TestValue;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.parser.AtomNodeDescription;
import pl.chyla.wyjec.parser.ExprTreePool;
import pl.chyla.wyjec.parser.ExpressionTreeDescription;
import pl.chyla.wyjec.parser.FunctionNodeDescription;
import pl.chyla.wyjec.parser.ValueNodeDescription;
import pl.chyla.wyjec.utils.ArrayUtils;


/**
 * Tests for {@link Evaluator} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestEvaluator extends TestCase {

    /**
     * Sample values of {@link Type#BOOLEAN} type.
     */
    private static final Value[] BOOLEAN_VALUES = {
        TestValue.makeBoolean(true), TestValue.makeBoolean(false)};

    /**
     * Sample values of {@link Type#INTEGER} type.
     */
    private static final Value[] INTEGER_VALUES = {
        TestValue.makeInteger(Long.MIN_VALUE), TestValue.makeInteger(0L),
        TestValue.makeInteger(123456L), TestValue.makeInteger(Long.MAX_VALUE)};

    /**
     * Sample values of {@link Type#STRING} type.
     */
    private static final Value[] STRING_VALUES = {
        TestValue.makeString(""), TestValue.makeString("x"), TestValue.makeString("abc")};

    /**
     * Sample values of {@link Type#FLOAT} type.
     */
    private static final Value[] FLOAT_VALUES = {
        TestValue.makeFloat(-Double.MIN_VALUE), TestValue.makeFloat(-Double.MAX_VALUE),
        TestValue.makeFloat(0.0), TestValue.makeFloat(1234.5),
        TestValue.makeFloat(Double.MIN_VALUE), TestValue.makeFloat(Double.MAX_VALUE)};

    /**
     * Sample values of {@link Type#DATE} type.
     */
    private static final Value[] DATE_VALUES = {
        TestValue.makeDate(1975, 3, 9, 16, 0, 0), TestValue.makeDate(1977, 2, 9, 21, 0, 0),
        TestValue.makeDate(1978, 8, 11, 15, 0, 0)};

    /**
     * All sample values of various types.
     */
    private static final Value[] ALL_VALUES =
        (Value[]) ArrayUtils.flattenArrays(
            new Value[][] {
                BOOLEAN_VALUES, INTEGER_VALUES, STRING_VALUES, FLOAT_VALUES, DATE_VALUES},
            new Value[0]);

    /**
     * A pool of expression tree objects used by the evaluator.
     */
    private ExprTreePool exprPool;

    /**
     * Evaluator instance used in the tests.
     */
    private Evaluator evaluator;

    /**
     * Empty default constructor
     */
    public TestEvaluator() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        Config config = Config.getDefaultConfig();
        exprPool = new ExprTreePool(1, config);
        evaluator = new Evaluator(config);
    }

    /**
     * Generates {@link ValueNodeDescription} instances for given values.
     *
     * @param values an array of values to convert to node descriptions
     * @return array of generated {@link ValueNodeDescription} instances
     */
    private ValueNodeDescription[] makeValuesNodesDescriptions(Value[] values) {
        ValueNodeDescription[] result = new ValueNodeDescription[values.length];
        for (int i = 0; i < values.length; i++) {
            result[i] = new ValueNodeDescription(values[i]);
        }
        return result;
    }

    /**
     * Tests the evaluator with a few expressions consisting of single constant values.
     */
    public void testSingleValue() throws EvaluationException {
        for (int i = 0; i < ALL_VALUES.length; i++) {
            Value value = ALL_VALUES[i];
            ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
                new ValueNodeDescription(value));
            exprPool.reset();
            Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
            assertTrue(value.equals(result));
        }
    }

    /**
     * Tests the evaluator with an expression consisting of single integer function with
     * constant values as parameters.
     */
    public void testIntegerFunction() throws EvaluationException {
        Value[] params = INTEGER_VALUES;
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnIntSum(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnIntSum.testExecute(params));
    }

    /**
     * Tests the evaluator with an expression consisting of single string function with
     * constant values as parameters.
     */
    public void testStringFunction() throws EvaluationException {
        Value[] params = STRING_VALUES;
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnStrConcat(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnStrConcat.testExecute(params));
    }

    /**
     * Tests the evaluator with an expression consisting of single integer function without
     * any parameters.
     */
    public void testIntegerFunctionNoParams() throws EvaluationException {
        Value[] params = new Value[0];
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnIntSum(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnIntSum.testExecute(params));
    }

    /**
     * Tests the evaluator with an expression consisting of single integer function with number
     * of parameters equal to soft limit.
     */
    public void testIntegerFunctionParamsSoftLimit() throws EvaluationException {
        int paramsSoftLimit = Config.getDefaultConfig().getFuncActualParamsSoftLimit();
        Value[] params = new Value[paramsSoftLimit];
        Arrays.fill(params, TestValue.makeInteger(1L));
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnIntSum(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnIntSum.testExecute(params));
    }

    /**
     * Tests the evaluator with an expression consisting of single integer function with number
     * of parameters exceeding soft limit.
     */
    public void testIntegerFunctionParamsOverSoftLimit() throws EvaluationException {
        int paramsSoftLimit = Config.getDefaultConfig().getFuncActualParamsSoftLimit();
        Value[] params = new Value[paramsSoftLimit + 1];
        Arrays.fill(params, TestValue.makeInteger(1L));
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnIntSum(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnIntSum.testExecute(params));
    }

    /**
     * Tests the evaluator with an expression consisting of nested functions.
     */
    public void testNestedFunctions() throws EvaluationException {
        Value[] aParams = INTEGER_VALUES;
        ValueNodeDescription[] aParamsDescs = makeValuesNodesDescriptions(aParams);
        Value[] bParams = new Value[Config.getDefaultConfig().getFuncActualParamsSoftLimit()];
        Arrays.fill(bParams, TestValue.makeInteger(1L));
        ValueNodeDescription[] bParamsDescs = makeValuesNodesDescriptions(bParams);
        Value cParam = TestValue.makeInteger(10L);
        ValueNodeDescription cParamDesc = new ValueNodeDescription(cParam);
        Function func = new FnIntSum();
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(func, new AtomNodeDescription[] {
                new FunctionNodeDescription(func, aParamsDescs),
                new FunctionNodeDescription(func, bParamsDescs),
                cParamDesc}));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        Value expResult = FnIntSum.testExecute(new Value[] {
            FnIntSum.testExecute(aParams), FnIntSum.testExecute(bParams), cParam});
        assertEquals(result, expResult);
    }

    /**
     * Tests the evaluator with an expression consisting of function supporting lazy evaluation
     * of parameters, using constant values as parameters.
     */
    public void testLazyEvaluationWithValues() throws EvaluationException {
        Value[] params = new Value[] {
            TestValue.makeInteger(1L), TestValue.makeInteger(5L), TestValue.makeInteger(10L)};
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnChoose(), makeValuesNodesDescriptions(params)));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnChoose.testExecute(params));
    }

    /**
     * Tests the evaluator with an expression consisting of function supporting lazy evaluation
     * of parameters, using other functions as parameters.
     */
    public void testLazyEvaluationWithFunctions() throws EvaluationException {
        Function chooseFunc = new FnChoose();
        FnFlag aFunc = new FnFlag();
        FnFlag bFunc = new FnFlag();
        AtomNodeDescription aFuncDescs =
            new FunctionNodeDescription(aFunc, new AtomNodeDescription[0]);
        AtomNodeDescription bFuncDescs =
            new FunctionNodeDescription(bFunc, new AtomNodeDescription[0]);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(chooseFunc, new AtomNodeDescription[] {
                new ValueNodeDescription(TestValue.makeInteger(0L)), aFuncDescs, bFuncDescs}));
        exprPool.reset();
        evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertTrue(aFunc.isEvaluated());
        assertFalse(bFunc.isEvaluated());
        aFunc.resetEvaluated();
        bFunc.resetEvaluated();
        exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(chooseFunc, new AtomNodeDescription[] {
                new ValueNodeDescription(TestValue.makeInteger(1L)), aFuncDescs, bFuncDescs}));
        exprPool.reset();
        evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertFalse(aFunc.isEvaluated());
        assertTrue(bFunc.isEvaluated());
    }

    /**
     * Tests the evaluator with an expression consisting of function supporting lazy evaluation
     * of parameters, using the same function for parameters to test function re-entrancy.
     */
    public void testLazyEvaluationWithSameFunction() throws EvaluationException {
        Function chooseFunc = new FnChoose();
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(chooseFunc, new AtomNodeDescription[] {
                new FunctionNodeDescription(chooseFunc, new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeInteger(1L)),
                    new ValueNodeDescription(TestValue.makeInteger(1L)),
                    new ValueNodeDescription(TestValue.makeInteger(0L))}),
                new FunctionNodeDescription(chooseFunc, new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeInteger(1L)),
                    new ValueNodeDescription(TestValue.makeInteger(5L)),
                    new ValueNodeDescription(TestValue.makeInteger(10L))}),
                new FunctionNodeDescription(chooseFunc, new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeInteger(1L)),
                    new ValueNodeDescription(TestValue.makeInteger(15L)),
                    new ValueNodeDescription(TestValue.makeInteger(20L))})}));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, TestValue.makeInteger(10L));
    }

    /**
     * Tests the evaluator with an expression consisting of function supporting lazy evaluation
     * of parameters, with number of parameters equal to soft limit.
     */
    public void testLazyEvaluationParamsSoftLimit() throws EvaluationException {
        int paramsSoftLimit = Config.getDefaultConfig().getFuncActualParamsSoftLimit();
        Value[] params = new Value[paramsSoftLimit];
        params[0] = TestValue.makeInteger(0L);
        Arrays.fill(params, 1, params.length, TestValue.makeInteger(1L));
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnChoose(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, TestValue.makeInteger(1L));
    }

    /**
     * Tests the evaluator with an expression consisting of function supporting lazy evaluation
     * of parameters, with number of parameters exceeding soft limit.
     */
    public void testLazyEvaluationParamsOverSoftLimit() throws EvaluationException {
        int paramsSoftLimit = Config.getDefaultConfig().getFuncActualParamsSoftLimit();
        Value[] params = new Value[paramsSoftLimit + 1];
        params[0] = TestValue.makeInteger(0L);
        Arrays.fill(params, 1, params.length, TestValue.makeInteger(1L));
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnChoose(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, TestValue.makeInteger(1L));
    }

    /**
     * Tests the evaluator with an expression consisting of function which throws an exception
     * during evaluation.
     */
    public void testFunctionException() {
        Value[] params = new Value[] {
            TestValue.makeInteger(5L), TestValue.makeInteger(1L)};
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnChoose(), paramsDescs));
        exprPool.reset();
        try {
            evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
            fail();
        } catch (EvaluationException e) {
            // OK
        }
    }

    /**
     * Tests the evaluator using {@link Config#getZeroMemLimitsConfig()} configuration.
     */
    public void testZeroMemLimitsConfig() throws EvaluationException {
        Config config = Config.getZeroMemLimitsConfig();
        exprPool = new ExprTreePool(1, config);
        evaluator = new Evaluator(config);
        Value[] params = STRING_VALUES;
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnStrConcat(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnStrConcat.testExecute(params));
    }

    /**
     * Tests the evaluator using very small (but big enough) value of
     * <code>StringLengthLimit</code> in the configuration.
     */
    public void testSmallStringLengthLimit() throws EvaluationException {
        Config config = Config.getDefaultConfig();
        config.setStringLengthLimit(3);
        exprPool = new ExprTreePool(1, config);
        evaluator = new Evaluator(config);
        Value[] params = {
            TestValue.makeString("a"), TestValue.makeString("b"), TestValue.makeString("c")};
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnStrConcat(), paramsDescs));
        exprPool.reset();
        Value result = evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
        assertEquals(result, FnStrConcat.testExecute(params));
    }

    /**
     * Tests the evaluator using very too small value of <code>StringLengthLimit</code> in the
     * configuration.
     */
    public void testTooSmallStringLengthLimit() {
        Config config = Config.getDefaultConfig();
        config.setStringLengthLimit(3);
        exprPool = new ExprTreePool(1, config);
        evaluator = new Evaluator(config);
        Value[] params = {
            TestValue.makeString("a"), TestValue.makeString("b"), TestValue.makeString("cd")};
        ValueNodeDescription[] paramsDescs = makeValuesNodesDescriptions(params);
        ExpressionTreeDescription exprTreeDesc = new ExpressionTreeDescription(
            new FunctionNodeDescription(new FnStrConcat(), paramsDescs));
        exprPool.reset();
        try {
            evaluator.evaluate(exprTreeDesc.makeExpressionTree(exprPool));
            fail();
        } catch (EvaluationException e) {
            // OK
        }
    }

}
