/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import java.util.Arrays;
import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.stdconsts.StandardConstantsParser;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.utils.RawString;


/**
 * Tests for {@link AtomNode} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestAtomNode extends TestCase {

    /**
     * Parser instance used for testing.
     */
    private Parser parser;

    /**
     * Empty default constructor
     */
    public TestAtomNode() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        parser = new Parser(
            Config.getDefaultConfig(), StandardFunctions.getFunctions(),
            StandardConstantsParser.getInstance());
    }

    /**
     * Checks recursively whether two given atom nodes are equivalent, i.e. represent the same
     * part of an expression tree.
     *
     * @param aAtomNode first atom node for comparison
     * @param bAtomNode second atom node for comparison
     */
    public static void checkAtomNodeEquivalence(AtomNode aAtomNode, AtomNode bAtomNode) {
        assertSame(aAtomNode.getAtomType(), bAtomNode.getAtomType());
        if (aAtomNode.getAtomType() == AtomType.FUNCTION) {
            FunctionNode aFuncNode = aAtomNode.getAsFunctionNode();
            FunctionNode bFuncNode = bAtomNode.getAsFunctionNode();
            Function aFunc = aFuncNode.getFunction();
            Function bFunc = bFuncNode.getFunction();
            assertEquals(aFunc.getName(), bFunc.getName());
            assertEquals(aFunc.getResultType(), bFunc.getResultType());
            Arrays.equals(
                aFunc.getSignature().getRequiredParametersTypes(),
                bFunc.getSignature().getRequiredParametersTypes());
            assertSame(
                aFunc.getSignature().getOptionalParametersType(),
                bFunc.getSignature().getOptionalParametersType());
            ParameterNode aParam = aFuncNode.getParameters();
            ParameterNode bParam = bFuncNode.getParameters();
            while (aParam != null && bParam != null) {
                checkAtomNodeEquivalence(aParam.getAtom(), bParam.getAtom());
                aParam = aParam.getNext();
                bParam = bParam.getNext();
            }
            assertTrue(aParam == null && bParam == null);
        } else {
            ValueNode aValueNode = aAtomNode.getAsValueNode();
            ValueNode bValueNode = bAtomNode.getAsValueNode();
            assertEquals(aValueNode.getValue(), bValueNode.getValue());
        }
    }

    /**
     * Recursively checks for given atom node and its sub-nodes if their <code>getAs*</code>
     * method behaves correctly.
     *
     * @param atomNode atom to check
     */
    private void checkAtomNodeGetAsResursive(AtomNode atomNode) {
        if (atomNode.getAtomType() == AtomType.FUNCTION) {
            try {
                atomNode.getAsValueNode();
                fail();
            } catch (UnsupportedOperationException e) {
                // OK
            }
            FunctionNode funcNode = atomNode.getAsFunctionNode();
            ParameterNode param = funcNode.getParameters();
            while (param != null) {
                checkAtomNodeGetAsResursive(param.getAtom());
                param = param.getNext();
            }
        } else {
            try {
                atomNode.getAsFunctionNode();
                fail();
            } catch (UnsupportedOperationException e) {
                // OK
            }
        }
    }

    /**
     * Tests whether {@link AtomNode#getAsFunctionNode()} and {@link AtomNode#getAsValueNode()}
     * methods behave correctly when called for invalid atom type.
     */
    public void testInvalidGetAs() throws InvalidInputException {
        ExpressionTree exprTree = parser.parse(
            new RawString("Test: {lcurly}{+|{2}|{3}}{rcurly}"));
        checkAtomNodeGetAsResursive(exprTree.getTopLevelAtom());
    }

}
