/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.BaseTestFunction;
import pl.chyla.wyjec.core.TestType;
import pl.chyla.wyjec.core.Type;


/**
 * Tests for {@link pl.chyla.wyjec.repo.EncodedTypeList} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestEncodedTypeList extends TestCase {

    /**
     * Maximum number of varying types used when generating variations of types sequence. When
     * generating such sequences, some number of constant types will be followed by up to
     * {@link #VARYING_TYPES_COUNT}-element variations of Wyjec types.
     */
    private static int VARYING_TYPES_COUNT = 4;

    /**
     * Empty default constructor
     */
    public TestEncodedTypeList() { }

    /**
     * Tests whether Wyjec implementation supports more than {@link #VARYING_TYPES_COUNT}
     * function parameters. Other tests in this class rely on this condition.
     */
    public void testSupportedParamsCount() {
        assertTrue(Config.FUNC_REQ_PARAMS_COUNT_LIMIT > VARYING_TYPES_COUNT);
    }

    /**
     * Tests {@link EncodedTypeList#isTypeListMatching(long, long)} method for a few sample
     * sequences of declared and actual types.
     */
    public void testIsTypeListMatchingSimple() {
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(new Type[] {}, Type.BOOLEAN),
                encodeTypeList(new Type[] {}, Type.NONE)));
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(new Type[] {}, Type.BOOLEAN),
                encodeTypeList(new Type[] {Type.BOOLEAN}, Type.NONE)));
        assertFalse(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(new Type[] {}, Type.BOOLEAN),
                encodeTypeList(new Type[] {Type.INTEGER}, Type.NONE)));
        assertFalse(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(
                    new Type[] {Type.DATE, Type.FLOAT, Type.BOOLEAN}, Type.INTEGER),
                encodeTypeList(new Type[] {Type.DATE, Type.FLOAT}, Type.NONE)));
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(
                    new Type[] {Type.DATE, Type.FLOAT, Type.BOOLEAN}, Type.INTEGER),
                encodeTypeList(
                    new Type[] {Type.DATE, Type.FLOAT, Type.BOOLEAN}, Type.NONE)));
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(
                    new Type[] {Type.DATE, Type.FLOAT, Type.BOOLEAN}, Type.INTEGER),
                encodeTypeList(
                    new Type[] {
                        Type.DATE, Type.FLOAT, Type.BOOLEAN, Type.INTEGER, Type.INTEGER},
                    Type.NONE)));
        assertFalse(
            EncodedTypeList.isTypeListMatching(
                encodeTypeList(
                    new Type[] {Type.DATE, Type.FLOAT, Type.BOOLEAN}, Type.INTEGER),
                encodeTypeList(
                    new Type[] {Type.DATE, Type.FLOAT, Type.BOOLEAN, Type.INTEGER, Type.DATE},
                    Type.NONE)));
    }

    /**
     * Tests {@link EncodedTypeList#isTypeListMatching(long, long)} method for a few sample
     * sequences of declared and actual types, passing partially used {@link Type} arrays to
     * {@link EncodedTypeList#encodeTypeList} method.
     */
    public void testIsTypeListMatchingPartialArray() {
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                EncodedTypeList.encodeTypeList(
                    new Type[] {Type.INTEGER, Type.BOOLEAN}, 0, Type.BOOLEAN),
                encodeTypeList(new Type[] {}, Type.NONE)));
        assertFalse(
            EncodedTypeList.isTypeListMatching(
                EncodedTypeList.encodeTypeList(
                    new Type[] {Type.INTEGER, Type.DATE}, 2, Type.BOOLEAN),
                encodeTypeList(new Type[] {Type.INTEGER, Type.BOOLEAN}, Type.NONE)));
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                EncodedTypeList.encodeTypeList(
                    new Type[] {Type.INTEGER, Type.DATE}, 1, Type.BOOLEAN),
                encodeTypeList(new Type[] {Type.INTEGER, Type.BOOLEAN}, Type.NONE)));
        assertTrue(
            EncodedTypeList.isTypeListMatching(
                EncodedTypeList.encodeTypeList(
                    new Type[] {Type.INTEGER, Type.DATE}, 2, Type.BOOLEAN),
                EncodedTypeList.encodeTypeList(
                    new Type[] {Type.INTEGER, Type.DATE, Type.FLOAT}, 2, Type.NONE)));
    }

    /**
     * Tests {@link EncodedTypeList#findMatchingTypeList(long[], long)} method for a few sample
     * sequences of declared and actual types.
     */
    public void testFindMatchingTypeListSimple() {
        // TODO
    }

    /**
     * Generates all possible variations of type declarations with up to
     * {@link #VARYING_TYPES_COUNT} required types and all possible variations of up to
     * {@link #VARYING_TYPES_COUNT} actual types, then verifies for all combinations of type
     * declarations and actual types if {@link EncodedTypeList#isTypeListMatching} method
     * returns valid results.
     */
    public void testShortParamsList() {
        TypesDeclaration[] typesDeclarationVariations = generateTypesDeclarationVariations(
            new Type[0], VARYING_TYPES_COUNT);
        Type[][] actualTypesVariations = generateActualTypesVariations(
            new Type[0], VARYING_TYPES_COUNT);
        for (int i = 0; i < typesDeclarationVariations.length; i++) {
            TypesDeclaration typesDeclaration = typesDeclarationVariations[i];
            long encDeclTypes = encodeTypeList(
                typesDeclaration.getRequiredTypes(), typesDeclaration.getOptionalType());
            for (int j = 0; j < actualTypesVariations.length; j++) {
                Type[] actualTypes = actualTypesVariations[j];
                long encActTypes = encodeTypeList(actualTypes, Type.NONE);
                boolean typesMatch = BaseTestFunction.checkParametersTypesMatch(
                    typesDeclaration.getRequiredTypes(), typesDeclaration.getOptionalType(),
                    actualTypes);
                assertEquals(
                    EncodedTypeList.isTypeListMatching(encDeclTypes, encActTypes), typesMatch);
            }
        }
    }

    /**
     * Generates all possible variations of type declarations with some number of identical
     * types followed by up to {@link #VARYING_TYPES_COUNT} varying types and all possible
     * variations of actual types with similar structure, then verifies for all combinations of
     * type declarations and actual types if {@link EncodedTypeList#isTypeListMatching} method
     * returns valid results.
     */
    public void testLongParamsList() {
        Type[] allRealTypes = TestType.getAllRealTypes();
        Type[] prefixTypes = new Type[Config.FUNC_REQ_PARAMS_COUNT_LIMIT - VARYING_TYPES_COUNT];
        for (int i = 0; i < allRealTypes.length; i++) {
            Arrays.fill(prefixTypes, allRealTypes[i]);
            TypesDeclaration[] typesDeclarationVariations = generateTypesDeclarationVariations(
                prefixTypes, VARYING_TYPES_COUNT);
            Type[][] actualTypesVariations = generateActualTypesVariations(
                prefixTypes, VARYING_TYPES_COUNT);
            for (int j = 0; j < typesDeclarationVariations.length; j++) {
                TypesDeclaration typesDeclaration = typesDeclarationVariations[j];
                long encDeclTypes = encodeTypeList(
                    typesDeclaration.getRequiredTypes(), typesDeclaration.getOptionalType());
                for (int k = 0; k < actualTypesVariations.length; k++) {
                    Type[] actualTypes = actualTypesVariations[k];
                    long encActTypes = encodeTypeList(actualTypes, Type.NONE);
                    boolean typesMatch = BaseTestFunction.checkParametersTypesMatch(
                        typesDeclaration.getRequiredTypes(), typesDeclaration.getOptionalType(),
                        actualTypes);
                    assertEquals(
                        EncodedTypeList.isTypeListMatching(encDeclTypes, encActTypes),
                        typesMatch);
                }
            }
        }
    }

    /**
     * Simple wrapper around {@link EncodedTypeList.encodeTypeList} that doesn't require
     * specifying the number of elements to use in <code>types</code> array (all elements of
     * this array will be used).
     *
     * @param types array of {@link Type} instances to encode
     * @param extType "extension" of sequence of types
     * @return
     *     representation of a sequence of types and an extension in a form of <code>long</code>
     *     value
     */
    private long encodeTypeList(Type[] types, Type extType) {
        return EncodedTypeList.encodeTypeList(types, types.length, extType);
    }

    /**
     * Generates arrays of actual types with each of them starting with given
     * <code>baseActualTypes</code> followed by variations (with repetitions) of all real Wyjec
     * types.
     *
     * @param baseActualTypes
     *     constant actual types at the beginning of required types sequence
     * @param maxExtraActualTypes
     *     maximum number of varying actual types put after the base ones (all possible
     *     variations with up to <code>maxExtraActualTypes</code> elements will be generated)
     * @return
     *     array of generated arrays of actual types
     */
    private Type[][] generateActualTypesVariations(
        Type[] baseActualTypes, int maxExtraActualTypes) {
        List actualTypesVariations = new LinkedList();
        Type[][] extraTypesSequences = generateTypesSequenceVariations(maxExtraActualTypes);
        for (int i = 0; i < extraTypesSequences.length; i++) {
            Type[] extraTypes = extraTypesSequences[i];
            Type[] actualTypes = new Type[baseActualTypes.length + extraTypes.length];
            System.arraycopy(baseActualTypes, 0, actualTypes, 0, baseActualTypes.length);
            System.arraycopy(
                extraTypes, 0, actualTypes, baseActualTypes.length, extraTypes.length);
            actualTypesVariations.add(actualTypes);
        }
        return (Type[][]) actualTypesVariations.toArray(new Type[0][]);
    }

    /**
     * Generates type declarations ({@link TypesDeclaration} instances) with a list of required
     * types starting with given <code>baseRequiredTypes</code> followed by variations (with
     * repetitions) of all real Wyjec types, combined with all possible Wyjec types (including
     * {@link Type#NONE}) as optional type.
     *
     * @param baseRequiredTypes
     *     constant required types at the beginning of required types sequence
     * @param maxExtraRequiredTypes
     *     maximum number of varying required types put after the base ones (all possible
     *     variations with up to <code>maxExtraRequiredTypes</code> elements will be generated)
     * @return
     *     array of generated type declarations
     */
    private TypesDeclaration[] generateTypesDeclarationVariations(
        Type[] baseRequiredTypes, int maxExtraRequiredTypes) {
        List typesDeclarations = new LinkedList();
        Type[][] extraTypesSequences = generateTypesSequenceVariations(maxExtraRequiredTypes);
        Type[] allTypes = TestType.getAllTypes();
        for (int i = 0; i < extraTypesSequences.length; i++) {
            Type[] extraTypes = extraTypesSequences[i];
            Type[] requiredTypes = new Type[baseRequiredTypes.length + extraTypes.length];
            System.arraycopy(baseRequiredTypes, 0, requiredTypes, 0, baseRequiredTypes.length);
            System.arraycopy(
                extraTypes, 0, requiredTypes, baseRequiredTypes.length, extraTypes.length);
            for (int j = 0; j < allTypes.length; j++) {
                Type optionalType = allTypes[j];
                typesDeclarations.add(new TypesDeclaration(requiredTypes, optionalType));
            }
        }
        return (TypesDeclaration[]) typesDeclarations.toArray(new TypesDeclaration[0]);
    }

    /**
     * Generates all variations (with repetitions) of all real types (with {@link Type#NONE}
     * excluded) with up to specified number of elements. Generated variations will have from
     * 0 to <code>maxTypesCount</code> elements.
     *
     * @param maxTypesCount maximum number of elements in generated variations
     * @return
     *     array of generated variations (elements of this array will be arrays containing from
     *     0 to <code>maxTypesCount</code> elements)
     */
    private Type[][] generateTypesSequenceVariations(int maxTypesCount) {
        List typesSequenceVariations = new LinkedList();
        Type[] allRealTypes = TestType.getAllRealTypes();
        for (int typesCount = 0; typesCount < maxTypesCount; typesCount++) {
            typesSequenceVariations.addAll(
                generatetVariations(allRealTypes, new Type[typesCount]));
        }
        return (Type[][]) typesSequenceVariations.toArray(new Type[0][]);
    }

    /**
     * Generates all n-element variations (with repetitions) of elements from given array.
     *
     * @param elements array containing elements to choose from when generating variations
     * @param selectionArray
     *     array defining number of elements in variation (array's length) and runtime type
     *     of generated arrays
     * @return
     *     list of arrays containing variations of given elements, runtime type of every list
     *     element will be the same as the type of <code>selectionArray</code>
     */
    private List generatetVariations(Object[] elements, Object[] selectionArray) {
        List variations = new LinkedList();
        int variationsCount = (int) Math.pow(elements.length, selectionArray.length);
        for (int i = 0; i < variationsCount; i++) {
            int index = i;
            Object[] variation = (Object[]) selectionArray.clone();
            for (int j = 0; j < selectionArray.length; j++) {
                variation[j] = elements[index % elements.length];
                index /= elements.length;
            }
            variations.add(variation);
        }
        return variations;
    }

}
