/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import junit.framework.TestCase;
import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.core.TestValue;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Tests for {@link pl.chyla.wyjec.stdconsts.StackedConstantsParser} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestStackedConstantsParser extends TestCase {

    /**
     * Empty default constructor
     */
    public TestStackedConstantsParser() { }

    /**
     * Checks whether given constants parser doesn't recognize any standard constants.
     *
     * @param parser parser to check
     */
    private void checkNoRecognizedStdConsts(IConstantsParser parser) {
        Value value = new Value(0);
        assertFalse(parser.parseConstant(value, new RawString("")));
        assertFalse(parser.parseConstant(value, new RawString(" ")));
        assertFalse(parser.parseConstant(value, new RawString("0")));
        assertFalse(parser.parseConstant(value, new RawString("1.0")));
        assertFalse(parser.parseConstant(value, new RawString("@1977-03-09")));
    }

    /**
     * Tests {@link StackedConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * the stack containing only empty parsers.
     */
    public void testParseConstantWithEmpty() {
        StackedConstantsParser parser = new StackedConstantsParser(new IConstantsParser[0]);
        checkNoRecognizedStdConsts(parser);
        parser = new StackedConstantsParser(
            new IConstantsParser[] {EmptyConstantsParser.getInstance()});
        checkNoRecognizedStdConsts(parser);
        parser = new StackedConstantsParser(
            new IConstantsParser[] {
                EmptyConstantsParser.getInstance(), EmptyConstantsParser.getInstance()});
        checkNoRecognizedStdConsts(parser);
    }

    /**
     * Tests {@link StackedConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * the stack containing two parsers: empty and custom.
     */
    public void testParseConstantWithEmptyAndNonEmpty() {
        Value value = new Value(0);
        StackedConstantsParser parser = new StackedConstantsParser(
            new IConstantsParser[] {
                EmptyConstantsParser.getInstance(), new XTrueConstantsParser()});
        checkNoRecognizedStdConsts(parser);
        assertTrue(parser.parseConstant(value, new RawString("x")));
        assertEquals(value, TestValue.makeBoolean(true));
        parser = new StackedConstantsParser(
            new IConstantsParser[] {
                new XTrueConstantsParser(), EmptyConstantsParser.getInstance()});
        checkNoRecognizedStdConsts(parser);
        assertTrue(parser.parseConstant(value, new RawString("x")));
        assertEquals(value, TestValue.makeBoolean(true));
    }

    /**
     * Tests {@link StackedConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * the stack containing three parsers: custom, empty and integer.
     */
    public void testParseConstantWithEmptyAndTwoNonEmpty() {
        Value value = new Value(0);
        StackedConstantsParser parser = new StackedConstantsParser(
            new IConstantsParser[] {
                new XTrueConstantsParser(), EmptyConstantsParser.getInstance(),
                IntegerConstantsParser.getInstance()});
        assertTrue(parser.parseConstant(value, new RawString("123")));
        assertEquals(value, TestValue.makeInteger(123L));
        assertTrue(parser.parseConstant(value, new RawString("x")));
        assertEquals(value, TestValue.makeBoolean(true));
        parser = new StackedConstantsParser(
            new IConstantsParser[] {
                IntegerConstantsParser.getInstance(), EmptyConstantsParser.getInstance(),
                new XTrueConstantsParser()});
        assertTrue(parser.parseConstant(value, new RawString("123")));
        assertEquals(value, TestValue.makeInteger(123L));
        assertTrue(parser.parseConstant(value, new RawString("x")));
        assertEquals(value, TestValue.makeBoolean(true));
    }

    /**
     * Tests {@link StackedConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * the stack containing an empty parser and two parsers that recognize the same name ("x").
     */
    public void testParseConstantWithEmptyAndTwoColliding() {
        Value value = new Value(0);
        StackedConstantsParser parser = new StackedConstantsParser(
            new IConstantsParser[] {
                new XTrueConstantsParser(), EmptyConstantsParser.getInstance(),
                new XFalseConstantsParser()});
        checkNoRecognizedStdConsts(parser);
        assertTrue(parser.parseConstant(value, new RawString("x")));
        assertEquals(value, TestValue.makeBoolean(true));
        parser = new StackedConstantsParser(
            new IConstantsParser[] {
                new XFalseConstantsParser(), EmptyConstantsParser.getInstance(),
                new XTrueConstantsParser()});
        checkNoRecognizedStdConsts(parser);
        assertTrue(parser.parseConstant(value, new RawString("x")));
        assertEquals(value, TestValue.makeBoolean(false));
    }


    /**
     * Constant parser that recognizes only "x" as a constant's name and maps it to boolean
     * true value.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private class XTrueConstantsParser implements IConstantsParser {

        /**
         * Parses given string as a representation of constant value, recognizing "x" as boolean
         * true value.
         *
         * @param result {@link Value} object used to store recognized constant value
         * @param constStr string representation of constant to parse
         */
        public boolean parseConstant(Value result, ReadableRawString constStr) {
            if (constStr.toString().equalsIgnoreCase("x")) {
                result.setBoolean(true);
                return true;
            } else {
                return false;
            }
        }

    }


    /**
     * Constant parser that recognizes only "x" as a constant's name and maps it to boolean
     * false value.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private class XFalseConstantsParser implements IConstantsParser {

        /**
         * Parses given string as a representation of constant value, recognizing "x" as boolean
         * false value.
         *
         * @param result {@link Value} object used to store recognized constant value
         * @param constStr string representation of constant to parse
         */
        public boolean parseConstant(Value result, ReadableRawString constStr) {
            if (constStr.toString().equalsIgnoreCase("x")) {
                result.setBoolean(false);
                return true;
            } else {
                return false;
            }
        }

    }

}
