/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.stringfuncs.ops;

import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.core.functions.BaseTestStringFunction;


/**
 * Tests for {@link FnLPad} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestFnLPad extends BaseTestStringFunction {

    /**
     * Constructs test instance with associated instance of the function.
     */
    public TestFnLPad() {
        super(new FnLPad());
    }

    /**
     * Evaluates {@link FnLPad} function with given parameters using different variants of
     * {@link pl.chyla.wyjec.utils.RawString} equivalents for <code>valueParam</code> and
     * <code>paddingParam</code> and checks if the result is always equal to the expected
     * result.
     *
     * @param valueParam value of <code>value</code> parameter to pass to the function
     * @param lengthParam value of <code>length</code> parameter to pass to the function
     * @param paddingParam value of <code>padding</code> parameter to pass to the function
     * @param expResult expected result of function evaluations
     */
    private void checkFnLPad(
        String valueParam, long lengthParam, String paddingParam, String expResult) {
        Value[] valueParamVariants = makeStringValueVariants(valueParam);
        Value[] paddingParamVariants = makeStringValueVariants(paddingParam);
        for (int i = 0; i < valueParamVariants.length; i++) {
            for (int j = 0; j < paddingParamVariants.length; j++) {
                Value[] params = {
                    valueParamVariants[i], makeIntegerValue(lengthParam),
                    paddingParamVariants[j]};
                checkFunctionResultVarSizes(params, expResult);
            }
        }
    }

    /**
     * Tests implementation of the function with empty "padding" parameter.
     */
    public void testEvaluateWithEmptyPadding() {
        String[] values = {"", "x", "abcde"};
        for (int i = 0; i < values.length; i++) {
            String value = values[i];
            checkFnLPad(value, Long.MIN_VALUE, "", value);
            checkFnLPad(value, -1, "", value);
            checkFnLPad(value, 0, "", value);
            checkFnLPad(value, value.length() - 1, "", value);
            checkFnLPad(value, value.length(), "", value);
            checkFnLPad(value, value.length() + 1, "", value);
        }
    }

    /**
     * Tests implementation of the function with "length" parameter less than or equal to the
     * length of the value.
     */
    public void testEvaluateShortLength() {
        String[] values = {"", "x", "abcde"};
        for (int i = 0; i < values.length; i++) {
            String value = values[i];
            checkFnLPad(value, Long.MIN_VALUE, ".-", value);
            checkFnLPad(value, -1, ".-", value);
            checkFnLPad(value, 0, ".-", value);
            checkFnLPad(value, value.length() - 1, ".-", value);
            checkFnLPad(value, value.length(), ".-", value);
        }
    }

    /**
     * Tests implementation of the function for various standard cases.
     */
    public void testEvaluate() {
        checkFnLPad("", 1, ".", ".");
        checkFnLPad("", 4, ".", "....");
        checkFnLPad("", 1, ".-", ".");
        checkFnLPad("", 2, ".-", ".-");
        checkFnLPad("", 3, ".-", ".-.");
        checkFnLPad("", 4, ".-", ".-.-");
        checkFnLPad("abc", 4, ".", ".abc");
        checkFnLPad("abc", 7, ".", "....abc");
        checkFnLPad("abc", 4, ".-", ".abc");
        checkFnLPad("abc", 5, ".-", ".-abc");
        checkFnLPad("abc", 6, ".-", ".-.abc");
        checkFnLPad("abc", 7, ".-", ".-.-abc");
        checkFnLPad("abc", 10, "12345", "1234512abc");
    }

}
