/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.testutils;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


/**
 * Simple wrapper around <code>Object</code> that implements <code>equals</code> and
 * <code>hashCode</code> methods as related to nested object but using their implementations
 * from <code>Object</code>. This allows e.g. for adding objects to collections using identity
 * comparison.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class IdEqualityWrapper {

    /**
     * The nested object.
     */
    private final Object obj;

    /**
     * Helper method for wrapping all objects in given list with <code>IdEqualityWrapper</code>.
     *
     * @param objects list of original objects
     * @return
     *     new list of <code>IdEqualityWrapper</code> instances that wrap the original objects
     */
    public static List wrapObjects(List objects) {
        List wrappedObjects = new ArrayList(objects.size());
        Iterator iter = objects.iterator();
        while (iter.hasNext()) {
            wrappedObjects.add(new IdEqualityWrapper(iter.next()));
        }
        return wrappedObjects;
    }

    /**
     * Constructs <code>IdEqualityWrapper</code> that wraps given object.
     *
     * @param obj object to wrap
     */
    public IdEqualityWrapper(Object obj) {
        this.obj = obj;
    }

    /**
     * Checks whether the wrapped object is identical to the one wrapped by another wrapper.
     *
     * @param other another wrapper to compare this one to
     * @return
     *     <code>true</code> if <code>other</code> is <code>IdEqualityWrapper</code> and its
     *     nested object is identical to the object nested in this wrapper; <code>false</code>
     *     otherwise
     */
    public boolean equals(Object other) {
        if (!(other instanceof IdEqualityWrapper)) {
            return false;
        }
        IdEqualityWrapper otherWrapper = (IdEqualityWrapper) other;
        return this.obj == otherWrapper.obj;
    }

    /**
     * Returns the default hash code (as implemented by <code>Object</code>) for the nested
     * object.
     */
    public int hashCode() {
        return System.identityHashCode(obj);
    }

    /**
     * Returns the object wrapped by this one.
     *
     * @return object nested in this wrapper
     */
    public Object getObject() {
        return obj;
    }

}
