/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Arrays;
import junit.framework.TestCase;


/**
 * Tests for {@link DoubleFormatter} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestDoubleFormatter extends TestCase {

    /**
     * Temporary buffer used for storing formatter's output during test execution
     */
    private static final char[] tmpBuf = new char[DoubleFormatter.MAX_CHARACTERS_COUNT];

    /**
     * Empty default constructor
     */
    public TestDoubleFormatter() { }

    /**
     * Tests {@link DoubleFormatter#formatToArray(int, char[], int)} method.
     * Calls the method with different buffer sizes, start positions and double values, parsing
     * formatter's output using <code>Double.parseDouble</code> and checking whether the result
     * is close enough to the original value.
     */
    public void testFormatToArrayAndGetCharactersCount() {
        double[] specialNumbers = {
            0.0, -102300000000004.56, 1e35, Double.NaN, Double.NEGATIVE_INFINITY,
            Double.POSITIVE_INFINITY, 12345.67890123456789012345, 12345.000890123456789012345,
            8.9234567823456171, -8.9234567823456171, -123456789012345678901.2345,
            -0.0000000000000001234567890123456789012345, -123443562445623453452E284, 2.0, 10.0,
            0.1, 0.100000000001, 0.1000000000056, 0.09999999999999999, -0.09999999999999999,
            123456789012345.6789, 123456789014.506789, 1234567890123456789e35, 45e20};
        for (int i = 0; i < specialNumbers.length; i++) {
            double value = specialNumbers[i];
            int charsCount = DoubleFormatter.formatToArray(value, tmpBuf, 0);
            assertTrue(charsCount > 0);
            String strValue = String.copyValueOf(tmpBuf, 0, charsCount);
            checkFormattedValue(value, strValue);
            for (int bufSize = 0; bufSize <= charsCount * 2; bufSize++) {
                char[] buf = new char[bufSize];
                for (int startPos = 0; startPos <= buf.length; startPos++) {
                    checkFormatter(value, strValue, buf, startPos, '\0');
                    checkFormatter(value, strValue, buf, startPos, ' ');
                }
            }
        }
    }

    /**
     * Tests the formatter against given value and output buffer, filling the whole buffer with
     * specified filler before the operation and checking whether characters before and after
     * formatters output hasn't been changed.
     *
     * @param value double value to format
     * @param strValue expected output of the formatter in case output buffer is big enough
     * @param buf output buffer for formatted value
     * @param startPos index of first buffer element to use for formatted value
     * @param filler character used to fill the whole buffer before launching formatter
     */
    private void checkFormatter(
        double value, String strValue, char[] buf, int startPos, char filler) {
        Arrays.fill(buf, filler);
        int charsCount = DoubleFormatter.formatToArray(value, buf, startPos);
        if (startPos + strValue.length() <= buf.length) {
            assertEquals(charsCount, strValue.length());
            assertEquals(String.copyValueOf(buf, startPos, charsCount), strValue);
            checkFiller(buf, 0, startPos, filler);
            checkFiller(buf, startPos + charsCount, buf.length, filler);
        } else {
            assertEquals(charsCount, 0);
            checkFiller(buf, 0, buf.length, filler);
        }
    }

    /**
     * Checks whether generated string form of double value matches given actual value.
     *
     * @param value double value passed to the formatter
     * @param strValue string form of the value generated by the formatter
     */
    private void checkFormattedValue(double value, String strValue) {
        if (strValue.equals("NaN")) {
            assertTrue(Double.isNaN(value));
        } else if (strValue.equals("inf")) {
            assertTrue(value == Double.POSITIVE_INFINITY);
        } else if (strValue.equals("-inf")) {
            assertTrue(value == Double.NEGATIVE_INFINITY);
        } else {
            double parsedValue = Double.parseDouble(strValue);
            if (Math.abs(value) < Double.MIN_VALUE) {
                assertTrue(Math.abs(parsedValue) < Double.MIN_VALUE);
            } else {
                assertEquals(1.0, parsedValue / value, 0.00000000000001);
            }
        }
    }

    /**
     * Checks whether specified range of components of given character array contains the same
     * given character.
     *
     * @param buf character buffer to check
     * @param startPos index of first buffer element to check
     * @param endPos index after the last buffer element to check
     * @param filler expected value of buffer elements
     */
    private void checkFiller(char[] buf, int startPos, int endPos, char filler) {
        for (int i = startPos; i < endPos; i++) {
            assertEquals(buf[i], filler);
        }
    }

}
