/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.EmptyStackException;
import junit.framework.TestCase;


/**
 * Tests for {@link IntStack} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestIntStack extends TestCase {

    /**
     * <code>IntStack</code> instance used for testing.
     */
    private IntStack stack;

    /**
     * Empty default constructor
     */
    public TestIntStack() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        stack = new IntStack();
    }

    /**
     * Generates short array containing a few sample integer numbers.
     *
     * @return
     */
    private int[] getSampleValues() {
        return new int[] {Integer.MIN_VALUE, -12345, 0, 54321, Integer.MAX_VALUE};
    }

    /**
     * Checks if a state of the stack is valid, assuming that given value is on its top. The
     * method uses {@link IntStack#isEmpty()}, {@link IntStack#peek()} and
     * {@link IntStack#pop()} methods.
     */
    private void checkWithTopValue(int expectedValue) {
        assertFalse(stack.isEmpty());
        assertEquals(stack.peek(), expectedValue);
    }

    /**
     * Checks if a state of the stack is valid, assuming that it's empty. The method uses
     * {@link IntStack#isEmpty()}, {@link IntStack#peek()} and {@link IntStack#pop()} methods.
     */
    private void checkWithEmptyStack() {
        assertTrue(stack.isEmpty());
        try {
            stack.peek();
            fail();
        } catch (EmptyStackException e) {
            // OK
        }
        try {
            stack.pop();
            fail();
        } catch (EmptyStackException e) {
            // OK
        }
    }

    /**
     * Performs a sequence of push + pop operations, verifying stack integrity after every
     * operation.
     */
    public void testPushPopDepth1() {
        int[] values = getSampleValues();
        for (int i = 0; i < values.length; i++) {
            int value = values[i];
            checkWithEmptyStack();
            stack.push(value);
            checkWithTopValue(value);
            assertEquals(stack.pop(), value);
        }
        checkWithEmptyStack();
    }

    /**
     * Puts a few numbers of the stack, performs a sequence of push + pop operations and pops
     * all numbers, verifying stack integrity after every operation.
     */
    public void testMultiplePushPop() {
        checkWithEmptyStack();
        int[] values = getSampleValues();
        for (int i = 0; i < values.length; i++) {
            int value = values[i];
            stack.push(value);
            checkWithTopValue(value);
            for (int j = 0; j < values.length; j++) {
                int topValue = values[j];
                stack.push(topValue);
                checkWithTopValue(topValue);
                assertEquals(stack.pop(), topValue);
            }
        }
        for (int i = values.length - 1; i >= 0; i--) {
            int value = values[i];
            checkWithTopValue(value);
            assertEquals(stack.pop(), value);
            for (int j = 0; j < values.length; j++) {
                int topValue = values[j];
                stack.push(topValue);
                checkWithTopValue(topValue);
                assertEquals(stack.pop(), topValue);
            }
        }
        checkWithEmptyStack();
    }

    /**
     * Tests push + pop operations on {@link IntStack} instance constructed with zero initial
     * capacity.
     */
    public void testZeroInitialCapacity() {
        stack = new IntStack(0);
        int[] values = getSampleValues();
        for (int i = 0; i < values.length; i++) {
            int value = values[i];
            checkWithEmptyStack();
            stack.push(value);
            checkWithTopValue(value);
            assertEquals(stack.pop(), value);
        }
        checkWithEmptyStack();
    }

    /**
     * Tests push + pop operations on {@link IntStack} instance constructed with non-zero
     * initial capacity.
     */
    public void testNonZeroInitialCapacity() {
        stack = new IntStack(1000);
        int[] values = getSampleValues();
        for (int i = 0; i < values.length; i++) {
            int value = values[i];
            checkWithEmptyStack();
            stack.push(value);
            checkWithTopValue(value);
            assertEquals(stack.pop(), value);
        }
        checkWithEmptyStack();
    }

    /**
     * Tests if the <code>IntStack</code> constructor properly notifies about invalid parameter.
     */
    public void testIntStackInvalidParameters() {
        try {
            new IntStack(-1);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

}
