/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import junit.framework.TestCase;


/**
 * Tests for {@link IntUtils} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestIntUtils extends TestCase {

    /**
     * Empty default constructor
     */
    public TestIntUtils() { }

    /**
     * Helper method checking whether selected bit in given number is set.
     *
     * @param value number to check for set bit
     * @param bitNumber position of bit to check (0 - 31)
     * @return true if <code>value</code> has bit at position <code>bitNumber</code> set,
     *         false otherwise
     */
    private boolean isBitSet(int value, int bitNumber) {
        return (value & (1 << bitNumber)) != 0;
    }

    /**
     * Tests {@link IntUtils#getWithHighestBitsSet(int)} method.
     * Passes all valid values of <code>bitNumber</code> (0 - 32) to the method and checks if
     * the highest <code>bitNumber</code> bits of returned number are set and all lower bits
     * are cleared.
     */
    public void testGetHighestBitsSet() {
        for (int i = 0; i <= 32; i++) {
            int value = IntUtils.getWithHighestBitsSet(i);
            for (int j = 0; j < 32 - i; j++) {
                assertFalse(isBitSet(value, j));
            }
            for (int j = 32 - i; j < 32; j++) {
                assertTrue(isBitSet(value, j));
            }
        }
    }

    /**
     * Tests {@link IntUtils#getLowestBitsSet(int)} method.
     * Passes all valid values of <code>bitNumber</code> (0 - 32) to the method and checks if
     * the lowest <code>bitNumber</code> bits of returned number are set and all higher bits
     * are cleared.
     */
    public void testWithGetLowestBitsSet() {
        for (int i = 0; i <= 32; i++) {
            int value = IntUtils.getLowestBitsSet(i);
            for (int j = 0; j < i; j++) {
                assertTrue(isBitSet(value, j));
            }
            for (int j = i; j < 32; j++) {
                assertFalse(isBitSet(value, j));
            }
        }
    }

    /**
     * Tests {@link IntUtils#isBetweenInclusive(int, int, int)} method.
     * Calls the method for selected values, including many border cases.
     */
    public void testIsBetweenInclusive() {
        int[] specialValues = {
            Integer.MIN_VALUE, -1000000, -1000, 0, 1000, 1000000, Integer.MAX_VALUE};
        for (int i = 0; i < specialValues.length; i++) {
            for (int j = i; j < specialValues.length; j++) {
                int min = specialValues[i];
                int max = specialValues[j];
                long longMin = min;
                long longMax = max;
                int mid = (int) ((longMin + longMax) / 2);
                for (int delta = -10; delta <= 10; delta++) {
                    int value1 = min + delta;
                    assertEquals(
                        IntUtils.isBetweenInclusive(value1, min, max),
                        value1 >= min && value1 <= max);
                    int value2 = max + delta;
                    assertEquals(
                        IntUtils.isBetweenInclusive(value2, min, max),
                        value2 >= min && value2 <= max);
                    int value3 = mid + delta;
                    assertEquals(
                        IntUtils.isBetweenInclusive(value3, min, max),
                        value3 >= min && value3 <= max);
                }
            }
        }
    }

    /**
     * Tests {@link IntUtils#compare(int, int)} method.
     * Calls the method for selected values, including many border cases.
     */
    public void testCompare() {
        int[] specialValues = {
            Integer.MIN_VALUE, -1000000, -1000, 0, 1000, 1000000, Integer.MAX_VALUE};
        for (int i = 0; i < specialValues.length; i++) {
            for (int j = 0; j < specialValues.length; j++) {
                for (int iDelta = -10; iDelta <= 10; iDelta++) {
                    for (int jDelta = -10; jDelta <= 10; jDelta++) {
                        int iValue = specialValues[i] + iDelta;
                        int jValue = specialValues[j] + jDelta;
                        if (iValue < jValue) {
                            assertTrue(IntUtils.compare(iValue, jValue) < 0);
                        } else if (iValue > jValue) {
                            assertTrue(IntUtils.compare(iValue, jValue) > 0);
                        } else {
                            assertEquals(IntUtils.compare(iValue, jValue), 0);
                        }
                    }
                }
            }
        }
    }

}
