/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import junit.framework.TestCase;


/**
 * Tests for {@link ReadableRawString} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestReadableRawString extends TestCase {

    /**
     * Empty default constructor
     */
    public TestReadableRawString() { }

    /**
     * Returns result of mathematical <i>signum</i> function for given value.
     *
     * @param value value to calculate signum for
     * @return
     *     signum for <code>value</code>, ie <code>-1</code>, <code>0</code> or <code>1</code>,
     *     depending on its sign
     */
    private int sgn(int value) {
        if (value > 0) {
            return 1;
        } else if (value < 0) {
            return -1;
        } else {
            return 0;
        }
    }

    /**
     * Tests {@link ReadableRawString#getStorage()} method.
     */
    public void testGetStorage() {
        char[] bufA = TestRawString.generateSampleString().toCharArray();
        char[] bufB = TestRawString.generateSampleString().toCharArray();
        RawString rstr = new RawString(bufA);
        assertSame(rstr.getStorage(), bufA);
        rstr.set(bufB);
        assertSame(rstr.getStorage(), bufB);
    }

    /**
     * Tests {@link ReadableRawString#getStartPos()}, {@link ReadableRawString#getEndPos()} and
     * {@link ReadableRawString#getLength()} methods.
     */
    public void testGetPositionsAndLength() {
        String str = TestRawString.generateSampleString();
        RawString rstr = new RawString(str);
        for (int startPos = 0; startPos <= str.length(); startPos++) {
            for (int endPos = startPos; endPos <= str.length(); endPos++) {
                rstr.setRange(startPos, endPos - startPos);
                assertEquals(rstr.getStartPos(), startPos);
                assertEquals(rstr.getEndPos(), endPos);
                assertEquals(rstr.getLength(), endPos - startPos);
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#toString()} method.
     */
    public void testToString() {
        String[] strings = TestRawString.generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            String str = strings[si];
            for (int startPos = 0; startPos <= str.length(); startPos++) {
                for (int lenth = 0; lenth <= str.length() - startPos; lenth++) {
                    RawString rstr = new RawString(str.toCharArray(), startPos, lenth);
                    assertEquals(rstr.toString(), str.substring(startPos, startPos + lenth));
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#equals(ReadableRawString)} method.
     */
    public void testEqualsReadableRawString() {
        String[] strings = TestRawString.generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            for (int sj = 0; sj < strings.length; sj++) {
                String strA = strings[si];
                String strB = strings[sj];
                RawString rstrA = new RawString(strA);
                RawString rstrB = new RawString(strB);
                assertEquals(
                    strA.equals(strB),
                    rstrA.equals(rstrB));
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#equals(ReadableRawString)} method checking whether it
     * obeys requirements defined in <code>Object.equals</code>.
     */
    public void testEqualsReadableRawStringRelationProperties() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstr = rawStrings[rsi];
            // equals(null) returns false
            assertFalse(rstr.equals(null));
        }
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstrA = rawStrings[rsi];
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                RawString rstrB = rawStrings[rsj];
                // symmetric
                assertEquals(
                    rstrA.equals(rstrB),
                    rstrA.equals(rstrB));
                // reflexive
                if (rstrA == rstrB) {
                    assertTrue(rstrA.equals(rstrB));
                }
                for (int rsk = 0; rsk < rawStrings.length; rsk++) {
                    RawString rstrC = rawStrings[rsk];
                    // transitive
                    if (rstrA.equals(rstrB) && rstrB.equals(rstrC)) {
                        assertTrue(rstrA.equals(rstrC));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#equals(Object)} method.
     */
    public void testEqualsObject() {
        String[] strings = TestRawString.generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            for (int sj = 0; sj < strings.length; sj++) {
                String strA = strings[si];
                String strB = strings[sj];
                Object rstrA = new RawString(strA);
                Object rstrB = new RawString(strB);
                assertEquals(
                    strA.equals(strB),
                    rstrA.equals(rstrB));
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#equals(Object)} method checking whether it obeys
     * requirements defined in <code>Object.equals</code>.
     */
    public void testEqualsObjectRelationProperties() {
        Object[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            Object rstr = rawStrings[rsi];
            // equals(null) returns false
            assertFalse(rstr.equals(null));
        }
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            Object rstrA = rawStrings[rsi];
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                Object rstrB = rawStrings[rsj];
                // symmetric
                assertEquals(
                    rstrA.equals(rstrB),
                    rstrA.equals(rstrB));
                // reflexive
                if (rstrA == rstrB) {
                    assertTrue(rstrA.equals(rstrB));
                }
                for (int rsk = 0; rsk < rawStrings.length; rsk++) {
                    Object rstrC = rawStrings[rsk];
                    // transitive
                    if (rstrA.equals(rstrB) && rstrB.equals(rstrC)) {
                        assertTrue(rstrA.equals(rstrC));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#equals(Object)} method using
     * non-<code>ReadableRawString</code> instances.
     */
    public void testEqualsObjectNotReadableRawString() {
        Object[] objects = {"", new Object(), new RawString[0], new ReadableRawString[0]};
        for (int oi = 0; oi < objects.length; oi++) {
            Object o = objects[oi];
            Object[] rawStrings = TestRawString.generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                Object rstr = rawStrings[rsi];
                assertFalse(rstr.equals(o));
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(ReadableRawString)} method.
     */
    public void testCompareToReadableRawString() {
        String[] strings = TestRawString.generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            for (int sj = 0; sj < strings.length; sj++) {
                String strA = strings[si];
                String strB = strings[sj];
                RawString rstrA = new RawString(strA);
                RawString rstrB = new RawString(strB);
                assertEquals(
                    sgn(strA.compareTo(strB)),
                    sgn(rstrA.compareTo(rstrB)));
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(ReadableRawString)} method checking whether it
     * obeys requirements defined in <code>Comparable.compareTo</code>.
     */
    public void testCompareToReadableRawStringRelationProperties() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstrA = rawStrings[rsi];
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                RawString rstrB = rawStrings[rsj];
                // sgn(a.compareTo(b)) == -sgn(b.compareTo(a))
                assertEquals(
                    sgn(rstrA.compareTo(rstrB)),
                    sgn(rstrA.compareTo(rstrB)));
                for (int rsk = 0; rsk < rawStrings.length; rsk++) {
                    RawString rstrC = rawStrings[rsk];
                    // transitive
                    if (sgn(rstrA.compareTo(rstrB)) == sgn(rstrB.compareTo(rstrC))) {
                        assertEquals(rstrA.compareTo(rstrC), sgn(rstrA.compareTo(rstrB)));
                    }
                    // a.compareTo(b) == 0 => sgn(a.compareTo(c)) == sgn(b.compareTo(c))
                    if (rstrA.compareTo(rstrB) == 0) {
                        assertEquals(sgn(rstrA.compareTo(rstrC)), sgn(rstrB.compareTo(rstrC)));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(Object)} method.
     */
    public void testCompareToObject() {
        String[] strings = TestRawString.generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            for (int sj = 0; sj < strings.length; sj++) {
                String strA = strings[si];
                String strB = strings[sj];
                Comparable rstrA = new RawString(strA);
                Comparable rstrB = new RawString(strB);
                assertEquals(
                    sgn(strA.compareTo(strB)),
                    sgn(rstrA.compareTo(rstrB)));
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(Object)} method method checking whether it
     * obeys requirements defined in <code>Comparable.compareTo</code>.
     */
    public void testCompareToObjectRelationProperties() {
        Comparable[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            Comparable rstrA = rawStrings[rsi];
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                Comparable rstrB = rawStrings[rsj];
                // sgn(a.compareTo(b)) == -sgn(b.compareTo(a))
                assertEquals(
                    sgn(rstrA.compareTo(rstrB)),
                    sgn(rstrA.compareTo(rstrB)));
                for (int rsk = 0; rsk < rawStrings.length; rsk++) {
                    Comparable rstrC = rawStrings[rsk];
                    // transitive
                    if (sgn(rstrA.compareTo(rstrB)) == sgn(rstrB.compareTo(rstrC))) {
                        assertEquals(rstrA.compareTo(rstrC), sgn(rstrA.compareTo(rstrB)));
                    }
                    // a.compareTo(b) == 0 => sgn(a.compareTo(c)) == sgn(b.compareTo(c))
                    if (rstrA.compareTo(rstrB) == 0) {
                        assertEquals(sgn(rstrA.compareTo(rstrC)), sgn(rstrB.compareTo(rstrC)));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(Object)} method using
     * non-<code>ReadableRawString</code> instances.
     */
    public void testCompareToObjectNotReadableRawString() {
        Object[] objects = {"", new Object(), new RawString[0], new ReadableRawString[0]};
        for (int oi = 0; oi < objects.length; oi++) {
            Object o = objects[oi];
            Comparable[] rawStrings = TestRawString.generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                Comparable rstr = rawStrings[rsi];
                try {
                    rstr.compareTo(o);
                    fail();
                } catch (ClassCastException e) {
                    // ok
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#hashCode()} method for different but equal strings.
     */
    public void testHashCodeAllEqual() {
        String[] strings = TestRawString.generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            String str = strings[si];
            RawString[] rawStrings =
                TestRawString.buildRawStringVariants(str);
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                    RawString rstrA = rawStrings[rsi];
                    RawString rstrB = rawStrings[rsj];
                    assertEquals(rstrA.hashCode(), rstrB.hashCode());
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#hashCode()} and
     * {@link ReadableRawString#equals(ReadableRawString)} methods, checking whether they return
     * consistent results.
     */
    public void testHashCodeWithEqualsReadableRawString() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                RawString rstrA = rawStrings[rsi];
                RawString rstrB = rawStrings[rsj];
                if (rstrA.equals(rstrB)) {
                    assertEquals(rstrA.hashCode(), rstrB.hashCode());
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#hashCode()} and
     * {@link ReadableRawString#equals(Object)} methods, checking whether they return consistent
     * results.
     */
    public void testHashCodeWithEqualsObject() {
        Comparable[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                Object rstrA = rawStrings[rsi];
                Object rstrB = rawStrings[rsj];
                if (rstrA.equals(rstrB)) {
                    assertEquals(rstrA.hashCode(), rstrB.hashCode());
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(ReadableRawString)} and
     * {@link ReadableRawString#equals(ReadableRawString)} methods, checking whether they return
     * consistent results.
     */
    public void testCompareToWithEqualsReadableRawString() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                RawString rstrA = rawStrings[rsi];
                RawString rstrB = rawStrings[rsj];
                assertEquals(rstrA.equals(rstrB), rstrA.compareTo(rstrB) == 0);
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#compareTo(Object)} and
     * {@link ReadableRawString#equals(Object)} methods, checking whether they return consistent
     * results.
     */
    public void testCompareToWithEqualsObject() {
        Comparable[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                Comparable rstrA = rawStrings[rsi];
                Comparable rstrB = rawStrings[rsj];
                assertEquals(rstrA.equals(rstrB), rstrA.compareTo(rstrB) == 0);
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#find(ReadableRawString, int)} method.
     */
    public void testFind() {
        checkFindWithSubstringAndIndex("AbcDefGhi", "D");
        checkFindWithSubstringAndIndex("AbcDefGhi", "Df");
        checkFindWithSubstringAndIndex("AbcDefGhi", "Abc");
        checkFindWithSubstringAndIndex("AbcDefGhi", "AbD");
        checkFindWithSubstringAndIndex("AbcDefGhi", "Def");
        checkFindWithSubstringAndIndex("AbcDefGhi", "DeG");
        checkFindWithSubstringAndIndex("AbcDefGhi", "Ghi");
        checkFindWithSubstringAndIndex("AbcDefGhi", "GhJ");
        checkFindWithSubstringAndIndex("AbcDefGhi", "GhiJ");
        checkFindWithSubstringAndIndex("AbcDefGhi", "AbcDefGhi");
        checkFindWithSubstringAndIndex("AbcDefGhi", "AbcDefGhiJ");
    }

    /**
     * Verifies whether {@link ReadableRawString#find(ReadableRawString, int)} method returns
     * correct results for invocations with different variations of given base string and
     * substring.
     *
     * @param str base string to search a substring in
     * @param subStr substring to search in the base string
     */
    private void checkFindWithSubstringAndIndex(String str, String subStr) {
        int subIndex = str.indexOf(subStr);
        RawString[] rawStringVariants = TestRawString.buildRawStringVariants(str);
        RawString[] rawSubStringVariants = TestRawString.buildRawStringVariants(subStr);
        for (int rsi = 0; rsi < rawStringVariants.length; rsi++) {
            RawString rstr = rawStringVariants[rsi];
            for (int rssi = 0; rssi < rawSubStringVariants.length; rssi++) {
                RawString rSubStr = rawSubStringVariants[rssi];
                for (int findStartPos = 0; findStartPos <= rstr.getLength(); findStartPos++) {
                    int idx = rstr.find(rSubStr, findStartPos);
                    if (findStartPos <= subIndex) {
                        assertEquals(idx, subIndex);
                    } else {
                        assertEquals(idx, -1);
                    }
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#find(ReadableRawString, int)} method, searching for empty
     * string.
     */
    public void testFindEmpty() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        RawString[] rawSubStringVariants = TestRawString.buildRawStringVariants("");
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstr = rawStrings[rsi];
            for (int rssi = 0; rssi < rawSubStringVariants.length; rssi++) {
                RawString rSubStr = rawSubStringVariants[rssi];
                for (int i = 0; i <= rstr.getLength(); i++) {
                    assertEquals(rstr.find(rSubStr, i), i);
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#find(ReadableRawString, int)} method with invalid
     * <code>findStartPos</code> argument.
     */
    public void testFindInvalidArgs() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        RawString[] rawSubStringVariants = TestRawString.buildRawStringVariants("abc");
        for (int rssi = 0; rssi < rawSubStringVariants.length; rssi++) {
            RawString rSubStr = rawSubStringVariants[rssi];
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                try {
                    rstr.find(rSubStr, -1);
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.find(rSubStr, rstr.getLength() + 1);
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
        }
    }

    /**
     * Verifies whether {@link ReadableRawString#find(char, int)} method returns correct results
     * for invocations with different variations of given base string and given character.
     *
     * @param str base string to search a character in
     * @param ch character to search in the base string
     */
    private void checkFindWithCharAndIndex(String str, char ch) {
        int chIndex = str.indexOf(ch);
        RawString[] rawStringVariants = TestRawString.buildRawStringVariants(str);
        for (int rsi = 0; rsi < rawStringVariants.length; rsi++) {
            RawString rstr = rawStringVariants[rsi];
            for (int findStartPos = 0; findStartPos <= rstr.getLength(); findStartPos++) {
                int idx = rstr.find(ch, findStartPos);
                if (findStartPos <= chIndex) {
                    assertEquals(idx, chIndex);
                } else {
                    assertEquals(idx, -1);
                }
            }
        }
    }

    /**
     * Tests {@link ReadableRawString#find(char, int)} method.
     */
    public void testFindChar() {
        checkFindWithCharAndIndex("AbcDefGhi", 'A');
        checkFindWithCharAndIndex("AbcDefGhi", 'b');
        checkFindWithCharAndIndex("AbcDefGhi", 'c');
        checkFindWithCharAndIndex("AbcDefGhi", 'G');
        checkFindWithCharAndIndex("AbcDefGhi", 'h');
        checkFindWithCharAndIndex("AbcDefGhi", 'i');
        checkFindWithCharAndIndex("", 'x');
        checkFindWithCharAndIndex("", ' ');
    }

    /**
     * Tests {@link ReadableRawString#find(char, int)} method with invalid
     * <code>findStartPos</code> argument.
     */
    public void testFindCharInvalidArgs() {
        RawString[] rawStrings = TestRawString.generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstr = rawStrings[rsi];
            try {
                rstr.find(' ', -1);
            } catch (IllegalArgumentException e) {
                // OK
            }
            try {
                rstr.find(' ', rstr.getLength() + 1);
            } catch (IllegalArgumentException e) {
                // OK
            }
        }
    }

}
